@@ -2,7 +2,7 @@
  * Test WebKDC user metadata retrieval.
  *
  * Written by Russ Allbery <rra@stanford.edu>
- * Copyright 2011
+ * Copyright 2011, 2012
  *     The Board of Trustees of the Leland Stanford Junior University
  *
  * See LICENSE for licensing terms.
@@ -51,16 +51,14 @@ test_validate(struct webauth_context *ctx, const char *code, bool success)
                       "...second is correct");
         }
         is_int(3, validate->loa, "...LoA is correct");
-    }        
-}    
+    }
+}
 
 
 int
 main(void)
 {
-    const char *principal;
-    char *keytab, *conf;
-    pid_t remctld;
+    struct kerberos_config *krbconf;
     struct webauth_context *ctx;
     struct webauth_user_config config;
     struct webauth_user_info *info;
@@ -68,32 +66,18 @@ main(void)
     struct webauth_login *login;
     int status;
 
-#ifndef HAVE_REMCTL
-    skip_all("built without remctl support");
-#endif
-#ifndef PATH_REMCTLD
-    skip_all("remctld not found");
-#endif
-    if (chdir(getenv("SOURCE")) < 0)
-        bail("can't chdir to SOURCE");
-    principal = kerberos_setup();
-    if (principal == NULL)
-        skip_all("Kerberos tests not configured");
-    keytab = test_file_path("data/test.keytab");
-    if (keytab == NULL)
-        skip_all("Kerberos tests not configured");
-
-    plan(58);
-
-    /* Set up the user metadata service configuration, testing error cases. */
+    /* Load test configuration and start remctl. */
+    krbconf = kerberos_setup(TAP_KRB_NEEDS_KEYTAB);
+    remctld_start(krbconf, "data/conf-webkdc", (char *) 0);
     if (webauth_context_init(&ctx, NULL) != WA_ERR_NONE)
         bail("cannot initialize WebAuth context");
-    conf = concatpath(getenv("SOURCE"), "data/conf-webkdc");
-    remctld = remctld_start(PATH_REMCTLD, principal, conf, NULL);
+
+    plan(91);
 
     /* Empty the KRB5CCNAME environment variable and make the library cope. */
     putenv((char *) "KRB5CCNAME=");
 
+    /* Set up the user metadata service configuration, testing error cases. */
     memset(&config, 0, sizeof(config));
     status = webauth_user_info(ctx, "test", "127.0.0.1", 0, &info);
     is_int(WA_ERR_INVALID, status, "Info without configuration");
@@ -112,14 +96,14 @@ main(void)
               webauth_error_message(ctx, status), "...with correct error");
     config.host = "localhost";
     config.port = 14373;
-    config.identity = principal;
+    config.identity = krbconf->principal;
     status = webauth_user_config(ctx, &config);
     is_int(WA_ERR_INVALID, status, "remctl config without keytab");
     is_string("invalid argument to function (keytab must be configured for"
               " remctl protocol)", webauth_error_message(ctx, status),
               "...with correct error");
-    config.keytab = keytab;
-    config.principal = principal;
+    config.keytab = krbconf->keytab;
+    config.principal = krbconf->principal;
     status = webauth_user_config(ctx, &config);
     is_int(WA_ERR_NONE, status, "Config with only host and protocol");
     status = webauth_user_info(ctx, "test", "127.0.0.1", 0, &info);
@@ -137,9 +121,10 @@ main(void)
     ok(info != NULL, "...info is not NULL");
     if (info == NULL) {
         is_string("", webauth_error_message(ctx, status), "...no error");
-        ok_block(16, 0, "...info is not NULL");
+        ok_block(17, 0, "...info is not NULL");
     } else {
         is_int(1, info->multifactor_required, "...multifactor required");
+        is_int(0, info->random_multifactor, "...random multifactor");
         is_int(3, info->max_loa, "...max LoA");
         is_int(1310675733, info->password_expires, "...password expires");
         ok(info->factors != NULL, "...factors is not NULL");
@@ -179,9 +164,25 @@ main(void)
     is_int(WA_ERR_NONE, status, "Metadata for mini succeeded");
     ok(info != NULL, "...mini is not NULL");
     if (info == NULL)
-        ok_block(5, 0, "Metadata failed");
+        ok_block(6, 0, "Metadata failed");
     else {
         is_int(0, info->multifactor_required, "...multifactor required");
+        is_int(0, info->random_multifactor, "...random multifactor");
+        is_int(1, info->max_loa, "...max LoA");
+        is_int(0, info->password_expires, "...password expires");
+        ok(info->factors == NULL, "...factors is NULL");
+        ok(info->logins == NULL, "...logins is NULL");
+    }
+
+    /* The same query, but with random multifactor. */
+    status = webauth_user_info(ctx, "mini", NULL, 1, &info);
+    is_int(WA_ERR_NONE, status, "Metadata for mini w/random succeeded");
+    ok(info != NULL, "...mini is not NULL");
+    if (info == NULL)
+        ok_block(6, 0, "Metadata failed");
+    else {
+        is_int(0, info->multifactor_required, "...multifactor required");
+        is_int(1, info->random_multifactor, "...random multifactor");
         is_int(1, info->max_loa, "...max LoA");
         is_int(0, info->password_expires, "...password expires");
         ok(info->factors == NULL, "...factors is NULL");
@@ -196,11 +197,65 @@ main(void)
 
     /* Attempt a login for a user who doesn't have multifactor configured. */
     status = webauth_user_validate(ctx, "mini", NULL, "123456", &validate);
-    is_int(status, WA_ERR_REMOTE_FAILURE, "Validate for invalid user fails");
+    is_int(WA_ERR_REMOTE_FAILURE, status, "Validate for invalid user fails");
     is_string("a remote service call failed (unknown user mini)",
               webauth_error_message(ctx, status), "...with correct error");
 
-    /* Clean up. */
-    remctld_stop(remctld);
+    /* Do a query for a user that should time out. */
+    config.timeout = 1;
+    status = webauth_user_config(ctx, &config);
+    is_int(WA_ERR_NONE, status, "Config with timeout");
+    status = webauth_user_info(ctx, "delay", NULL, 0, &info);
+    is_int(WA_ERR_REMOTE_FAILURE, status, "Metadata for delay fails");
+    is_string("a remote service call failed"
+              " (error receiving token: timed out)",
+              webauth_error_message(ctx, status), "...with correct error");
+
+    /* Attempt a login for a user that should time out. */
+    status = webauth_user_validate(ctx, "delay", NULL, "123456", &validate);
+    is_int(WA_ERR_REMOTE_FAILURE, status, "Validate for delay fails");
+    is_string("a remote service call failed"
+              " (error receiving token: timed out)",
+              webauth_error_message(ctx, status), "...with correct error");
+
+    /* Try the query again with ignore_failure set. */
+    config.ignore_failure = 1;
+    status = webauth_user_config(ctx, &config);
+    is_int(WA_ERR_NONE, status, "Config with timeout and ignore failure");
+    status = webauth_user_info(ctx, "delay", NULL, 0, &info);
+    is_int(status, WA_ERR_NONE, "Metadata for delay now succeeds");
+    if (info == NULL)
+        ok_block(6, 0, "Metadata failed");
+    else {
+        is_int(0, info->multifactor_required, "...multifactor required");
+        is_int(0, info->random_multifactor, "...random multifactor");
+        is_int(0, info->max_loa, "...max LoA");
+        is_int(0, info->password_expires, "...password expires");
+        ok(info->factors == NULL, "...factors is NULL");
+        ok(info->logins == NULL, "...logins is NULL");
+    }
+
+    /* Try the query again with ignore_failure and random multifactor. */
+    is_int(WA_ERR_NONE, status, "Config with timeout, ignore, random");
+    status = webauth_user_info(ctx, "delay", NULL, 1, &info);
+    is_int(status, WA_ERR_NONE, "Metadata for delay w/random succeeds");
+    if (info == NULL)
+        ok_block(6, 0, "Metadata failed");
+    else {
+        is_int(0, info->multifactor_required, "...multifactor required");
+        is_int(0, info->random_multifactor, "...random multifactor");
+        is_int(0, info->max_loa, "...max LoA");
+        is_int(0, info->password_expires, "...password expires");
+        ok(info->factors == NULL, "...factors is NULL");
+        ok(info->logins == NULL, "...logins is NULL");
+    }
+
+    /* Attempt a login again, which should still fail. */
+    status = webauth_user_validate(ctx, "delay", NULL, "123456", &validate);
+    is_int(WA_ERR_REMOTE_FAILURE, status, "Validate for delay fails");
+    is_string("a remote service call failed"
+              " (error receiving token: timed out)",
+              webauth_error_message(ctx, status), "...with correct error");
+
     return 0;
 }
